#define _XOPEN_SOURCE
#include <Xm/Xm.h>
#include <Xm/MainW.h>
#include <Xm/Scale.h>
#include <Xm/LabelG.h>
#include <Xm/PushB.h>
#include <Xm/MessageB.h>
#include <Xm/RowColumn.h>
#include <Xm/Separator.h>

#include <Xmd/Icon.h>
#include <Xmd/Exec.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/wait.h>

#include "icons/level0.xbm"
#include "icons/level1.xbm"
#include "icons/level2.xbm"
#include "icons/level3.xbm"
#include "icons/level4.xbm"
#include "icons/charging.xbm"
#include "icons/error.xbm"
#include "icons/not.xbm"

typedef enum {
	BatteryStateCharging,
	BatteryStateNotCharging,
	BatteryStateFull,
	BatteryStateDischarging,
	BatteryStateError
} BatteryState;

typedef struct {
	int          level;
	BatteryState state;
	int          hours;
	int          minutes;
	int          seconds;
} BatteryInfo;

BatteryInfo getBatteryInfo (void);
Pixmap      selectBatteryIcon (BatteryInfo, Pixmap[8]);
void        loadAllPixmaps (Widget, Pixmap[8]);
void        batteryPoll (XtPointer, XtIntervalId *);
void        resetBatteryPollTimeout(void);
void        batteryInfoDialog (Widget, XtPointer, XtPointer);

Pixmap levels[8]     = { 0 };
Pixmap iconLevels[8] = { 0 };
XtAppContext application;
Widget icon;
Widget text;
Widget layout;
Widget topLevel;

int main (int argc, char *argv[]) {
	topLevel = XtVaAppInitialize (
		&application, "Battery",
		NULL, 0,
		&argc, argv,
		NULL,
		XmNtitle, "Battery",
		XmNiconName, "Battery",
		NULL);
	Widget window = XtVaCreateManagedWidget (
		"window", xmMainWindowWidgetClass, topLevel,
		NULL);
	layout = XtVaCreateWidget (
		"layout", xmRowColumnWidgetClass, window,
		XmNorientation, XmHORIZONTAL,
		NULL);
	
	loadAllPixmaps(topLevel, iconLevels);
	loadAllPixmaps(layout, levels);
	icon = XtVaCreateManagedWidget (
		"batteryIcon", xmPushButtonWidgetClass, layout,
		XmNleftAttachment,  XmATTACH_FORM,
		XmNlabelType,       XmPIXMAP,
		NULL);
	XtAddCallback(icon, XmNactivateCallback, batteryInfoDialog, NULL);
	
	XtVaCreateManagedWidget (
		"separator", xmSeparatorWidgetClass, layout,
		XmNorientation, XmVERTICAL,
		NULL);
	
	text = XtVaCreateManagedWidget (
		"text", xmLabelGadgetClass, layout,
		XmNalignment,       XmALIGNMENT_CENTER,
		NULL);
		
	batteryPoll(NULL, NULL);

	XtManageChild(layout);
	XtRealizeWidget(topLevel);
	XtAppMainLoop(application);
}

BatteryInfo getBatteryInfo (void) {
	BatteryInfo result = { 0 };
	char charging[16]  = { 0 };
	int battery;

	pid_t child;
	FILE *stream = XmdVaPipedExecPath (
		"acpi", &child, "r",
		"acpi", "-b",
		NULL);
	if (stream == NULL) {
		result.state = BatteryStateError;
		return result;
	}
	fscanf (
		stream, "Battery %d: %16s %d%%, %d:%d:%d",
		&battery, charging, &result.level,
		&result.hours, &result.minutes, &result.seconds);
	
	switch (charging[0]) {
	case 'C': result.state = BatteryStateCharging;    break;
	case 'N': result.state = BatteryStateNotCharging; break;
	case 'D': result.state = BatteryStateDischarging; break;
	case 'F': result.state = BatteryStateFull;        break;
	default: result.state = BatteryStateError;        break;
	}
	if (fclose(stream) != 0) {
		result.state = BatteryStateError;
	}
	waitpid(child, NULL, 0);
	return result;
}

Pixmap selectBatteryIcon (BatteryInfo info, Pixmap levels[8]) {
	if (info.state == BatteryStateCharging) {
		return levels[5];
	} else if (info.state == BatteryStateNotCharging) {
		return levels[7];
	} else if (info.state == BatteryStateError) {
		return levels[6];
	} else if (info.level < 15) {
		return levels[0];
	} else if (info.level < 35) {
		return levels[1];
	} else if (info.level < 65) {
		return levels[2];
	} else if (info.level < 85) {
		return levels[3];
	} else {
		return levels[4];
	}
}

void loadAllPixmaps (Widget widget, Pixmap destination[8]) {
	destination[0] = XmdLoadBitmapIcon(widget, level0);
	destination[1] = XmdLoadBitmapIcon(widget, level1);
	destination[2] = XmdLoadBitmapIcon(widget, level2);
	destination[3] = XmdLoadBitmapIcon(widget, level3);
	destination[4] = XmdLoadBitmapIcon(widget, level4);
	destination[5] = XmdLoadBitmapIcon(widget, charging);
	destination[6] = XmdLoadBitmapIcon(widget, error);
	destination[7] = XmdLoadBitmapIcon(widget, not);
}

void batteryPoll (XtPointer clientData, XtIntervalId *timer) {
	(void)(clientData);
	(void)(timer);
	BatteryInfo info = getBatteryInfo();
	
	XtVaSetValues (
		icon,
		XmNlabelType,   XmPIXMAP,
		XmNlabelPixmap, selectBatteryIcon(info, levels),
		NULL);
	XtVaSetValues (
		topLevel,
		XmNiconPixmap, selectBatteryIcon(info, iconLevels),
		NULL);
	
	char buffer[32];
	snprintf(buffer, XtNumber(buffer), "%d%%", info.level);
	
	/* if we don't unmanage and then re-manage the child, the text flies to
	 * the top which is rather annoying. this little quirk bent me over and
	 * fucked me for hours. */
	XtUnmanageChild(text);
	XmString string = XmStringCreateLocalized(buffer);
	XtVaSetValues (
		text,
		XmNlabelString, string,
		NULL);
	XmStringFree(string);
	XtManageChild(text);
	
	XtVaSetValues (
		topLevel,
		XmNiconName, buffer,
		NULL);
	
	resetBatteryPollTimeout();
}

void resetBatteryPollTimeout (void) {
	XtAppAddTimeOut(application, 2000, batteryPoll, NULL);
}

void batteryInfoDialog (Widget button, XtPointer clientData, XtPointer callData) {
	(void)(clientData);
	(void)(callData);
	
	BatteryInfo info = getBatteryInfo();
	
	static const String states[] = {
		"Charging",
		"Not charging",
		"Full",
		"Discharging",
		"Error"
	};
	
	String messageBuffer = NULL;
	switch (info.state) {
	case BatteryStateCharging:
		XtAsprintf (
			&messageBuffer,
			"%d%%, %s\n"
			"%d:%02d:%02d until charged.",
			info.level, states[info.state],
			info.hours, info.minutes, info.seconds);
		break;
	case BatteryStateDischarging:
		XtAsprintf (
			&messageBuffer,
			"%d%%, %s\n"
			"%d:%02d:%02d until empty.",
			info.level, states[info.state],
			info.hours, info.minutes, info.seconds);
		break;
		
	default:
		XtAsprintf (
			&messageBuffer,
			"%d%%, %s",
			info.level, states[info.state]);
		break;
	}

	Arg args[5] = { 0 };
	int n = 0;
	XmString message = XmStringCreateLocalized(messageBuffer);
	XtSetArg(args[n], XmNmessageString, message); n ++;
	XmString title = XmStringCreateLocalized("Battery Status");
	XtSetArg(args[n], XmNdialogTitle, title); n ++;
	
	Widget dialog = XmCreateInformationDialog(button, "batteryInfo", args, n);
	
	XmStringFree(message);
	XmStringFree(title);
	
	XtManageChild(dialog);
	XtPopup(XtParent(dialog), XtGrabNone);
}
